@@ -1,40 +1,41 @@
-// Copyright 2009-2021 Intel Corporation
+// Copyright 2018 Intel Corporation
 // SPDX-License-Identifier: Apache-2.0
 
+#include "common/common.h"
+#include "common/timer.h"
+#include "utils/arg_parser.h"
+#include "utils/image_io.h"
+#include "utils/device_info.h"
 #include <iostream>
 #include <fstream>
+#include <iomanip>
 #include <cassert>
 #include <limits>
 #include <cmath>
 #include <signal.h>
-
 #ifdef VTUNE
 #include <ittnotify.h>
 #endif
 
-#include <OpenImageDenoise/oidn.hpp>
-
-#include "common/timer.h"
-#include "apps/utils/image_io.h"
-#include "apps/utils/arg_parser.h"
-
 OIDN_NAMESPACE_USING
-using namespace oidn;
 
 void printUsage()
 {
   std::cout << "Intel(R) Open Image Denoise" << std::endl;
-  std::cout << "usage: oidnDenoise [-d/--device default|cpu]" << std::endl
+  std::cout << "usage: oidnDenoise [-d/--device [0-9]+|default|cpu|sycl|cuda|hip|metal]" << std::endl
             << "                   [-f/--filter RT|RTLightmap]" << std::endl
             << "                   [--hdr color.pfm] [--ldr color.pfm] [--srgb] [--dir directional.pfm]" << std::endl
             << "                   [--alb albedo.pfm] [--nrm normal.pfm] [--clean_aux]" << std::endl
             << "                   [--is/--input_scale value]" << std::endl
-            << "                   [-o/--output output.pfm] [-r/--ref reference_output.pfm]" << std::endl
+            << "                   [-o/--output output.pfm]" << std::endl
+            << "                   [-r/--ref reference_output.pfm] [--maxerror e]" << std::endl
             << "                   [-t/--type float|half]" << std::endl
+            << "                   [-q/--quality default|h|high|b|balanced|f|fast]" << std::endl
             << "                   [-w/--weights weights.tza]" << std::endl
             << "                   [--threads n] [--affinity 0|1] [--maxmem MB] [--inplace]" << std::endl
-            << "                   [--bench ntimes] [-v/--verbose 0-3]" << std::endl
-            << "                   [-h/--help]" << std::endl;
+            << "                   [--buffer host|device|managed]" << std::endl
+            << "                   [-n times_to_run] [-v/--verbose 0-3]" << std::endl
+            << "                   [--ld|--list_devices] [-h/--help]" << std::endl;
 }
 
 void errorCallback(void* userPtr, Error error, const char* message)
@@ -51,48 +52,47 @@ void signalHandler(int signal)
 
 bool progressCallback(void* userPtr, double n)
 {
-  if (isCancelled)
-  {
-    std::cout << std::endl;
-    return false;
-  }
-  std::cout << "\rDenoising " << int(n * 100.) << "%" << std::flush;
-  return true;
+  std::cout << "\rDenoising " << int(n * 100.) << "% " << std::flush;
+  return !isCancelled;
 }
 
 std::vector<char> loadFile(const std::string& filename)
 {
   std::ifstream file(filename, std::ios::binary);
   if (file.fail())
-    throw std::runtime_error("cannot open file: " + filename);
+    throw std::runtime_error("cannot open file: '" + filename + "'");
   file.seekg(0, file.end);
   const size_t size = file.tellg();
   file.seekg(0, file.beg);
   std::vector<char> buffer(size);
   file.read(buffer.data(), size);
   if (file.fail())
-    throw std::runtime_error("error reading from file");
+    throw std::runtime_error("error reading from file: '" + filename + "'");
   return buffer;
 }
 
 int main(int argc, char* argv[])
 {
   DeviceType deviceType = DeviceType::Default;
+  PhysicalDeviceRef physicalDevice;
   std::string filterType = "RT";
   std::string colorFilename, albedoFilename, normalFilename;
   std::string outputFilename, refFilename;
   std::string weightsFilename;
+  Quality quality = Quality::Default;
+  Storage bufferStorage = Storage::Undefined;
   bool hdr = false;
   bool srgb = false;
   bool directional = false;
   float inputScale = std::numeric_limits<float>::quiet_NaN();
   bool cleanAux = false;
-  Format dataType = Format::Undefined;
-  int numBenchmarkRuns = 0;
+  DataType dataType = DataType::Void;
+  int numRuns = 1;
   int numThreads = -1;
   int setAffinity = -1;
   int maxMemoryMB = -1;
   bool inplace = false;
+  double errorThreshold = -1;
   int verbose = -1;
 
   // Parse the arguments
@@ -110,13 +110,11 @@ int main(int argc, char* argv[])
       std::string opt = args.getNextOpt();
       if (opt == "d" || opt == "dev" || opt == "device")
       {
-        const auto val = args.getNextValue();
-        if (val == "default" || val == "Default")
-          deviceType = DeviceType::Default;
-        else if (val == "cpu" || val == "CPU")
-          deviceType = DeviceType::CPU;
+        std::string value = args.getNext();
+        if (isdigit(value[0]))
+          physicalDevice = fromString<int>(value);
         else
-          throw std::invalid_argument("invalid device");
+          deviceType = fromString<DeviceType>(value);
       }
       else if (opt == "f" || opt == "filter")
         filterType = args.getNextValue();
@@ -145,47 +143,74 @@ int main(int argc, char* argv[])
         outputFilename = args.getNextValue();
       else if (opt == "r" || opt == "ref" || opt == "reference")
         refFilename = args.getNextValue();
-      else if (opt == "is" || opt == "input_scale" || opt == "inputScale" || opt == "inputscale")
-        inputScale = args.getNextValueFloat();
-      else if (opt == "clean_aux" || opt == "cleanAux")
+      else if (opt == "is" || opt == "input_scale" || opt == "input-scale" || opt == "inputScale" || opt == "inputscale")
+        inputScale = args.getNextValue<float>();
+      else if (opt == "clean_aux" || opt == "clean-aux" || opt == "cleanAux" || opt == "cleanaux")
         cleanAux = true;
       else if (opt == "t" || opt == "type")
       {
-        const auto val = args.getNextValue();
-        if (val == "f" || val == "float" || val == "Float" || val == "fp32")
-          dataType = Format::Float;
-        else if (val == "h" || val == "half" || val == "Half" || val == "fp16")
-          dataType = Format::Half;
+        const auto val = toLower(args.getNextValue());
+        if (val == "f" || val == "float" || val == "fp32")
+          dataType = DataType::Float32;
+        else if (val == "h" || val == "half" || val == "fp16")
+          dataType = DataType::Float16;
         else
           throw std::runtime_error("invalid data type");
       }
+      else if (opt == "q" || opt == "quality")
+      {
+        const auto val = toLower(args.getNextValue());
+        if (val == "default")
+          quality = Quality::Default;
+        else if (val == "h" || val == "high")
+          quality = Quality::High;
+        else if (val == "b" || val == "balanced")
+          quality = Quality::Balanced;
+        else if (val == "f" || val == "fast")
+          quality = Quality::Fast;
+        else
+          throw std::runtime_error("invalid filter quality mode");
+      }
       else if (opt == "w" || opt == "weights")
         weightsFilename = args.getNextValue();
-      else if (opt == "bench" || opt == "benchmark")
-        numBenchmarkRuns = std::max(args.getNextValueInt(), 0);
+      else if (opt == "n")
+        numRuns = std::max(args.getNextValue<int>(), 1);
       else if (opt == "threads")
-        numThreads = args.getNextValueInt();
+        numThreads = args.getNextValue<int>();
       else if (opt == "affinity")
-        setAffinity = args.getNextValueInt();
+        setAffinity = args.getNextValue<int>();
       else if (opt == "maxmem" || opt == "maxMemoryMB")
-        maxMemoryMB = args.getNextValueInt();
+        maxMemoryMB = args.getNextValue<int>();
       else if (opt == "inplace")
         inplace = true;
+      else if (opt == "buffer")
+      {
+        const auto val = toLower(args.getNextValue());
+        if (val == "host")
+          bufferStorage = Storage::Host;
+        else if (val == "device")
+          bufferStorage = Storage::Device;
+        else if (val == "managed")
+          bufferStorage = Storage::Managed;
+        else
+          throw std::runtime_error("invalid storage mode");
+      }
+      else if (opt == "maxerror" || opt == "maxError" || opt == "max_error" || opt == "maxerr")
+        errorThreshold = args.getNextValue<double>();
       else if (opt == "v" || opt == "verbose")
-        verbose = args.getNextValueInt();
+        verbose = args.getNextValue<int>();
+      else if (opt == "ld" || opt == "list_devices" || opt == "list-devices" || opt == "listDevices" || opt == "listdevices")
+        return printPhysicalDevices();
       else if (opt == "h" || opt == "help")
       {
         printUsage();
         return 1;
       }
       else
-        throw std::invalid_argument("invalid argument");
+        throw std::invalid_argument("invalid argument '" + opt + "'");
     }
 
-    if (!refFilename.empty() && numBenchmarkRuns > 0)
-      throw std::runtime_error("reference and benchmark modes cannot be enabled at the same time");
-
-  #if defined(OIDN_X64)
+  #if defined(OIDN_ARCH_X64)
     // Set MXCSR flags
     if (!refFilename.empty())
     {
@@ -205,7 +230,11 @@ int main(int argc, char* argv[])
     std::cout << "Initializing device" << std::endl;
     Timer timer;
 
-    DeviceRef device = newDevice(deviceType);
+    DeviceRef device;
+    if (physicalDevice)
+      device = physicalDevice.newDevice();
+    else
+      device = newDevice(deviceType);
 
     const char* errorMessage;
     if (device.getError(errorMessage) != Error::None)
@@ -222,11 +251,12 @@ int main(int argc, char* argv[])
 
     const double deviceInitTime = timer.query();
 
+    deviceType = device.get<DeviceType>("type");
     const int versionMajor = device.get<int>("versionMajor");
     const int versionMinor = device.get<int>("versionMinor");
     const int versionPatch = device.get<int>("versionPatch");
 
-    std::cout << "  device=" << (deviceType == DeviceType::Default ? "default" : (deviceType == DeviceType::CPU ? "CPU" : "unknown"))
+    std::cout << "  device=" << deviceType
               << ", version=" << versionMajor << "." << versionMinor << "." << versionPatch
               << ", msec=" << (1000. * deviceInitTime) << std::endl;
 
@@ -237,26 +267,35 @@ int main(int argc, char* argv[])
     std::cout << "Loading input" << std::endl;
 
     if (!albedoFilename.empty())
-      input = albedo = loadImage(device, albedoFilename, 3, false, dataType);
+    {
+      input = albedo = loadImage(device, albedoFilename, false, dataType, bufferStorage);
+      albedo->toDevice();
+    }
 
     if (!normalFilename.empty())
-      input = normal = loadImage(device, normalFilename, 3, dataType);
+    {
+      input = normal = loadImage(device, normalFilename, dataType, bufferStorage);
+      normal->toDevice();
+    }
 
     if (!colorFilename.empty())
-      input = color = loadImage(device, colorFilename, 3, srgb, dataType);
+    {
+      input = color = loadImage(device, colorFilename, srgb, dataType, bufferStorage);
+      color->toDevice();
+    }
 
     if (!input)
       throw std::runtime_error("no input image specified");
 
     if (!refFilename.empty())
     {
-      ref = loadImage(device, refFilename, 3, srgb, dataType);
-      if (ref->dims() != input->dims())
+      ref = loadImage(device, refFilename, srgb, dataType);
+      if (ref->getDims() != input->getDims())
         throw std::runtime_error("invalid reference output image");
     }
 
-    const int width  = input->width;
-    const int height = input->height;
+    const int width  = input->getW();
+    const int height = input->getH();
     std::cout << "Resolution: " << width << "x" << height << std::endl;
 
     // Initialize the output image
@@ -264,7 +303,14 @@ int main(int argc, char* argv[])
     if (inplace)
       output = input;
     else
-      output = std::make_shared<ImageBuffer>(device, width, height, 3, input->dataType);
+    {
+      output = std::make_shared<ImageBuffer>(device, width, height, input->getC(),
+                                             input->getDataType(), bufferStorage);
+    }
+
+    std::shared_ptr<ImageBuffer> inputCopy;
+    if (inplace && numRuns > 1)
+      inputCopy = input->clone();
 
     // Load the filter weights if specified
     std::vector<char> weights;
@@ -281,13 +327,13 @@ int main(int argc, char* argv[])
     FilterRef filter = device.newFilter(filterType.c_str());
 
     if (color)
-      filter.setImage("color", color->data(), color->format(), color->width, color->height);
+      filter.setImage("color", color->getBuffer(), color->getFormat(), color->getW(), color->getH());
     if (albedo)
-      filter.setImage("albedo", albedo->data(), albedo->format(), albedo->width, albedo->height);
+      filter.setImage("albedo", albedo->getBuffer(), albedo->getFormat(), albedo->getW(), albedo->getH());
     if (normal)
-      filter.setImage("normal", normal->data(), normal->format(), normal->width, normal->height);
+      filter.setImage("normal", normal->getBuffer(), normal->getFormat(), normal->getW(), normal->getH());
 
-    filter.setImage("output", output->data(), output->format(), output->width, output->height);
+    filter.setImage("output", output->getBuffer(), output->getFormat(), output->getW(), output->getH());
 
     if (filterType == "RT")
     {
@@ -308,13 +354,16 @@ int main(int argc, char* argv[])
     if (cleanAux)
       filter.set("cleanAux", cleanAux);
 
+    if (quality != Quality::Default)
+      filter.set("quality", quality);
+
     if (maxMemoryMB >= 0)
       filter.set("maxMemoryMB", maxMemoryMB);
 
     if (!weights.empty())
       filter.setData("weights", weights.data(), weights.size());
 
-    const bool showProgress = !ref && numBenchmarkRuns == 0 && verbose <= 2;
+    const bool showProgress = verbose <= 1;
     if (showProgress)
     {
       filter.setProgressMonitorFunction(progressCallback);
@@ -329,17 +378,72 @@ int main(int argc, char* argv[])
               << ", msec=" << (1000. * filterInitTime) << std::endl;
 
     // Denoise the image
-    if (!showProgress)
-      std::cout << "Denoising" << std::endl;
-    timer.reset();
+    uint32_t prevHash = 0;
+    for (int run = 0; run < numRuns; ++run)
+    {
+      if (inplace && run > 0)
+        input->write(0, input->getByteSize(), inputCopy->getHostData());
 
-    filter.execute();
+      if (!showProgress)
+        std::cout << "Denoising" << std::endl;
+      timer.reset();
 
-    const double denoiseTime = timer.query();
-    if (showProgress)
-      std::cout << std::endl;
-    if (verbose <= 2)
-      std::cout << "  msec=" << (1000. * denoiseTime) << std::endl;
+      filter.execute();
+
+      const double denoiseTime = timer.query();
+
+      if (showProgress)
+        std::cout << std::endl;
+      std::cout << "  msec=" << (1000. * denoiseTime);
+
+      output->toHost();
+
+      if (numRuns > 1 || verbose >= 2)
+      {
+        // Compute a hash of the output
+        const size_t numBytes = output->getByteSize();
+        const uint8_t* outputBytes = static_cast<const uint8_t*>(output->getHostData());
+        uint32_t hash = 0x811c9dc5;
+        for (size_t i = 0; i < numBytes; ++i)
+        {
+          hash ^= outputBytes[i];
+          hash *= 0x1000193;
+        }
+        std::cout << ", hash=" << std::hex << std::setfill('0') << std::setw(8) << hash << std::dec << std::endl;
+
+        if (run > 0 && hash != prevHash)
+          throw std::runtime_error("output hash mismatch (non-deterministic output)");
+        prevHash = hash;
+      }
+      else
+        std::cout << std::endl;
+
+      if (run == 0 && ref)
+      {
+        // Verify the output values
+        std::cout << "Verifying output" << std::endl;
+
+        if (errorThreshold < 0.)
+          errorThreshold = (input == normal || directional) ? 0.05 : 0.003;
+        size_t numErrors;
+        double avgError;
+        std::tie(numErrors, avgError) = compareImage(*output, *ref, errorThreshold);
+
+        std::cout << "  values=" << output->getSize()
+                  << ", errors=" << numErrors << ", avgerror=" << avgError << std::endl;
+
+        if (numErrors > 0)
+        {
+          // Save debug images
+          std::cout << "Saving debug images" << std::endl;
+          saveImage("denoise_in.pfm",  *input,  srgb);
+          saveImage("denoise_out.pfm", *output, srgb);
+          saveImage("denoise_ref.pfm", *ref,    srgb);
+
+          throw std::runtime_error("output does not match the reference");
+        }
+      }
+    }
 
     if (showProgress)
     {
@@ -353,53 +457,8 @@ int main(int argc, char* argv[])
       std::cout << "Saving output" << std::endl;
       saveImage(outputFilename, *output, srgb);
     }
-
-    if (ref)
-    {
-      // Verify the output values
-      std::cout << "Verifying output" << std::endl;
-
-      const float threshold = (output->dataType == Format::Float) ? 1e-4f : 1e-2f;
-      size_t numErrors;
-      float maxError;
-      std::tie(numErrors, maxError) = compareImage(*output, *ref, threshold);
-
-      std::cout << "  values=" << output->size() << ", errors=" << numErrors << ", maxerror=" << maxError << std::endl;
-
-      if (numErrors > 0)
-      {
-        // Save debug images
-        std::cout << "Saving debug images" << std::endl;
-        saveImage("denoise_in.ppm",  *input,  srgb);
-        saveImage("denoise_out.ppm", *output, srgb);
-        saveImage("denoise_ref.ppm", *ref,    srgb);
-
-        throw std::runtime_error("output does not match the reference");
-      }
-    }
-
-    if (numBenchmarkRuns > 0)
-    {
-      // Benchmark loop
-    #ifdef VTUNE
-      __itt_resume();
-    #endif
-
-      std::cout << "Benchmarking: " << "ntimes=" << numBenchmarkRuns << std::endl;
-      timer.reset();
-
-      for (int i = 0; i < numBenchmarkRuns; ++i)
-        filter.execute();
-
-      const double totalTime = timer.query();
-      std::cout << "  sec=" << totalTime << ", msec/image=" << (1000.*totalTime / numBenchmarkRuns) << std::endl;
-
-    #ifdef VTUNE
-      __itt_pause();
-    #endif
-    }
   }
-  catch (std::exception& e)
+  catch (const std::exception& e)
   {
     std::cerr << "Error: " << e.what() << std::endl;
     return 1;
