@@ -1,11 +1,11 @@
-// Copyright 2009-2021 Intel Corporation
+// Copyright 2020 Intel Corporation
 // SPDX-License-Identifier: Apache-2.0
 
 #pragma once
 
-#include "common/math.h"
+#include "common/platform.h"
 
-namespace oidn {
+OIDN_NAMESPACE_BEGIN
 
   // Simple and very fast LCG random number generator
   class Random
@@ -14,37 +14,37 @@ namespace oidn {
     uint32_t state;
 
   public:
-    __forceinline Random(uint32_t seed = 1) : state(seed) {}
+    oidn_inline Random(uint32_t seed = 1) : state(seed) {}
 
-    __forceinline void reset(uint32_t seed = 1)
+    oidn_inline void reset(uint32_t seed = 1)
     {
       state = (seed * 8191) ^ 140167;
     }
 
-    __forceinline void next()
+    oidn_inline void next()
     {
       const uint32_t multiplier = 1664525;
       const uint32_t increment  = 1013904223;
       state = multiplier * state + increment;
     }
 
-    __forceinline uint32_t get1ui()
+    oidn_inline uint32_t getUInt()
     {
       next();
       return state;
     }
 
-    __forceinline int get1i()
+    oidn_inline int getInt()
     {
       next();
       return state;
     }
 
-    __forceinline float get1f()
+    oidn_inline float getFloat()
     {
       next();
-      return toFloatUnorm(state);
+      return float(state) * 2.3283064365386962890625e-10f; // x / 2^32
     }
   };
 
-} // namespace oidn
\ No newline at end of file
+OIDN_NAMESPACE_END
\ No newline at end of file
